#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@File    :   performance.py    
@Contact :   wangcc@csc.com.cn
@License :   (C)Copyright 2017-2018, CSC

@Modify Time      @Author    @Version    @Desciption
------------      -------    --------    -----------
2021/11/3 18:14   wangcc     1.0         None
'''


import pandas as pd
import numpy as np
from pandas import DataFrame
from datetime import date, datetime
pd.options.mode.chained_assignment = None


def calculate_statistics(df: DataFrame = None, prin_output=False, output=print, risk_free=0.023, perf_cal_return="gross return"):
    """"""
    # Check for init DataFrame
    # default perf_cal_return = "gross return"
    if (df is None) | (df.empty):
        # Set all statistics to 0 if no trade.
        start_date = pd.to_datetime(datetime.now())
        end_date = pd.to_datetime(datetime.now())
        total_days = 0
        profit_days = 0
        loss_days = 0
        end_balance = 0
        max_drawdown = 0
        max_drawdown_duration = 0
        total_commission = 0
        daily_commission = 0
        total_turnover = 0
        daily_turnover = 0
        total_return = 0
        annual_return = 0
        daily_return = 0
        return_std = 0
        sharpe_ratio = 0
        return_drawdown_ratio = 0
        ic = 0
        Rankic = 0
        net_return = 0
        return_turnover_ratio = 0
    else:
        # Calculate balance related time series data
        df.loc[:, ("balance")] = df[perf_cal_return].cumsum() + 1

        # When balance falls below 0, set daily return to 0

        df.loc[:, ("highlevel")] = (
            df["balance"].rolling(
                min_periods=1, window=len(df), center=False).max()
        )
        df.loc[:, ("drawdown")] = df["balance"] - df["highlevel"]

        # Calculate statistics value
        start_date = df.index[0]
        end_date = df.index[-1]

        total_days = len(df)
        profit_days = len(df[df[perf_cal_return] > 0])
        loss_days = len(df[df[perf_cal_return] < 0])

        end_balance = df["balance"].iloc[-1]
        max_drawdown = df["drawdown"].min()
        max_drawdown_end = df["drawdown"].idxmin()

        if isinstance(max_drawdown_end, date):
            max_drawdown_start = df["balance"][:max_drawdown_end].idxmax()
            max_drawdown_duration = (max_drawdown_end - max_drawdown_start).days
        else:
            max_drawdown_duration = 0

        total_commission = df["commission"].sum()
        daily_commission = total_commission / total_days

        total_turnover = df["turnover"].sum()
        daily_turnover = total_turnover / total_days

        total_return = df[perf_cal_return].sum()
        daily_return = df[perf_cal_return].mean() * 100
        return_std = df[perf_cal_return].std() * 100
        annual_return = daily_return * 250
        net_return = df["return"].mean() * 250 * 100

        daily_risk_free = risk_free / 240  # np.sqrt(240)
        sharpe_ratio = (daily_return - daily_risk_free) / max(return_std, 0.0001) * np.sqrt(240)

        return_drawdown_ratio = -total_return / max_drawdown
        return_turnover_ratio = daily_return / daily_turnover * 100

        if "ic" in df.columns:
            ic = df["ic"].mean()
        else:
            ic = np.nan
        if "Rankic" in df.columns:
            Rankic = df["Rankic"].mean()
        else:
            Rankic = np.nan

    # Output
    if prin_output:
        output("-" * 30)

        output(f"日均收益率：\t{daily_return:,.2f}%")
        output(f"收益标准差：\t{return_std:,.2f}%")
        output(f"Sharpe Ratio：\t{sharpe_ratio:,.2f}")
        output(f"收益回撤比：\t{return_drawdown_ratio:,.2f}")
        output(f"bp收益换手比：\t{return_turnover_ratio:,.2f}")
        output(f"净年化收益率: \t{net_return}:,.2f")

        output(f"IC：\t{ic:,.2f}")
        output(f"RankIC：\t{Rankic:,.2f}")

        output(f"首个交易日：\t{start_date}")
        output(f"最后交易日：\t{end_date}")

        output(f"总交易日：\t{total_days}")
        output(f"盈利交易日：\t{profit_days}")
        output(f"亏损交易日：\t{loss_days}")

        output(f"起始资金：\t{1:,.2f}")
        output(f"结束资金：\t{end_balance:,.2f}")

        output(f"总收益率：\t{total_return:,.2f}%")
        output(f"年化收益：\t{annual_return:,.2f}%")
        output(f"最大回撤: \t{max_drawdown:,.2f}")
        output(f"最长回撤天数: \t{max_drawdown_duration}")

        output(f"总手续费：\t{total_commission:,.2f}")
        output(f"总换手%：\t{total_turnover:,.2f}")

        output(f"日均手续费：\t{daily_commission:,.2f}")
        output(f"日均换手%：\t{daily_turnover:,.2f}")

    statistics = {
        "total_return%": total_return * 100,
        "annual_return%": annual_return,
        "daily_return%": daily_return,
        "return_std%": return_std,
        "sharpe_ratio": sharpe_ratio,
        "ic": ic,
        "Rankic": Rankic,
        "net_annual_return%": net_return,
        "total_commission": total_commission,
        "daily_commission": daily_commission,
        "total_turnover%": total_turnover * 100,
        "daily_turnover%": daily_turnover * 100,
        "max_drawdown": max_drawdown,
        "max_drawdown_duration": max_drawdown_duration,
        "return_drawdown_ratio": return_drawdown_ratio,
        "bpMargin": return_turnover_ratio,
        "start_date": start_date,
        "end_date": end_date,
        "total_days": total_days,
        "profit_days": profit_days,
        "loss_days": loss_days,
        "capital": 1,
        "end_balance": end_balance,
    }

    # Filter potential error infinite value
    for key, value in statistics.items():
        if value in (np.inf, -np.inf):
            value = 0
        statistics[key] = np.nan_to_num(value)

    groupretre = df[df.columns[df.columns.str.startswith('group_')]]
    if not groupretre.empty:
        g_retr_dict = groupretre.sum(axis=0).to_dict()
        if prin_output:
            for key, value in g_retr_dict.items():
                output(key, f": \t{value:,.2f}")
        statistics = {**statistics, **g_retr_dict}

    return pd.Series(statistics)


#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@File    :   __init__.py.py    
@Contact :   wangcc@csc.com.cn
@License :   (C)Copyright 2017-2018, CSC

@Modify Time      @Author    @Version    @Desciption
------------      -------    --------    -----------
2021/11/1 19:02   wangcc     1.0         None
'''

# import lib
from .object import context
from .engine import BackTestingEngine


#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@File    :   lib.py    
@Contact :   wangcc@csc.com.cn
@License :   (C)Copyright 2017-2018, CSC

@Modify Time      @Author    @Version    @Desciption
------------      -------    --------    -----------
2021/11/3 13:26   wangcc     1.0         None
'''

import sys

sys.path.append("..")
import numpy as np
import numba as nb
import pandas as pd
import bottleneck as bn
from .data_utils import ED
from datetime import datetime
from time import time
from numpy.linalg import LinAlgError
from scipy.stats import rankdata


def indu_neutral(values, indu):
    """ minus industry mean"""
    tmp = np.full_like(values, np.nan)
    for i, (y, dummy) in enumerate(zip(values, indu)):
        dummy[np.isnan(dummy)] = 9988877
        mask = ~np.isnan(y)
        y, dummy = y[mask], dummy[mask]
        if y.size > 0:
            uniq, indices = np.unique(dummy, return_inverse=True)
            groups = np.full((uniq.shape[0], indices.shape[0]), np.nan)
            arrs = np.arange(indices.shape[0])
            groups[indices, arrs] = y
            tmp[i, mask] = y - np.nanmean(groups, axis=1)[indices]
    return tmp


def mv_neutral(mat, sqmv):
    """ get residual after regression with sqmv """
    tmp = np.full_like(mat, np.nan)
    for i, (y, x) in enumerate(zip(mat, sqmv)):
        mask = ~(np.isnan(y) | np.isnan(x))
        y, x = y[None, mask], x[None, mask]
        x = np.concatenate([x, np.ones((1, x.shape[1]))], axis=0)
        try:
            tmp[i, mask] = reg_residual(y, x)
        except np.linalg.LinAlgError:
            tmp[i] = np.full_like(mask, np.nan)
    return mat


def indu_encode(indu):
    """ indu: 1d array """
    uniq, indices = np.unique(indu, return_inverse=True)
    groups = np.full((uniq.size, indices.size), 0.)
    arrs = np.arange(indices.size)
    groups[indices, arrs] = 1.
    return groups.astype(np.float32)


@nb.jit(nopython=True)
def reg_residual(y_arr, x_arr):
    """get residual after regression with market value
    :param y_arr:
    :param x_arr:
    :return:
    """
    if y_arr.size == 0:
        return y_arr
    beta = np.linalg.inv(x_arr @ x_arr.T) @ x_arr @ y_arr.T
    return y_arr - beta.T @ x_arr


def wash_outliers(mat: np.ndarray, num_sigma: int):
    """清除outlier"""
    mean_arr = np.expand_dims(bn.nanmean(mat, axis=1), axis=1)
    std_arr = np.expand_dims(bn.nanstd(mat, axis=1), axis=1)
    outlier_mask = np.abs(mat - mean_arr) > mean_arr + std_arr * num_sigma

    mat_copy = mat.copy()
    mat_copy[outlier_mask] = np.nan
    mean_arr_cp = np.expand_dims(bn.nanmean(mat, axis=1), axis=1)
    std_arr_cp = np.expand_dims(bn.nanstd(mat, axis=1), axis=1)
    mat = np.clip(mat, mean_arr_cp - std_arr_cp * num_sigma, mean_arr_cp + std_arr_cp * num_sigma)
    return mat


def winsorize(mat: np.ndarray, dnlimit=0.025, uplimit=0.975, axis=1) -> np.ndarray:
    """ 将极值替换分位点数 """
    dnq = np.nanquantile(mat, dnlimit, axis=axis, keepdims=True)
    upq = np.nanquantile(mat, uplimit, axis=axis, keepdims=True)
    mat = np.clip(mat, dnq, upq)
    return mat


def standardize(mat):
    """ minus mean divide std"""
    mean = np.expand_dims(bn.nanmean(mat, axis=1), axis=1)
    stdn = np.expand_dims(bn.nanstd(mat, axis=1), axis=1)
    equal = stdn == 0
    stdn[equal] = 1
    return (mat - mean) / stdn


def mkt_neutral(factor):
    """ 返回多空booksize中性(值中性）"""
    alpha = factor - np.nanmean(factor, axis=1, keepdims=True)
    pos_mask, neg_mask = alpha > 0.000000001, alpha < -0.000000001
    zero_mask = ~(pos_mask | neg_mask | (np.isnan(alpha)))
    alpha = alpha * pos_mask / np.expand_dims(bn.nansum(alpha * pos_mask, axis=1), axis=1) \
        - alpha * neg_mask / np.expand_dims(bn.nansum(alpha * neg_mask, axis=1), axis=1)
    alpha[np.isinf(alpha)] = np.nan
    alpha[zero_mask] = 0
    alpha *= 0.5
    return alpha


def weight_control(FactorMonlix, OpParaPercent=0.05, OpParaNumIteration4Truncate=4):
    TotalNumRow, TotalNumCol = FactorMonlix.shape
    #
    if np.array([OpParaPercent]).shape.__len__() == 1:
        assert OpParaPercent > 0
        OpParaPercent = np.ones((TotalNumRow, TotalNumCol), dtype=float) * OpParaPercent
    else:
        assert OpParaPercent.shape == (TotalNumRow, TotalNumCol)
    #
    OpParaPercent[np.isnan(FactorMonlix)] = np.nan
    #
    assert OpParaNumIteration4Truncate > 0
    #
    OpedFactorMonlix = np.empty_like(FactorMonlix) * np.nan
    id4allok = True
    #
    for i in range(TotalNumRow):
        temp_vector = FactorMonlix[i, :]
        temp_ceil_vector = OpParaPercent[i, :]
        temp_notnan_vector = ~np.isnan(temp_vector)
        if not any(temp_notnan_vector):
            continue
        #
        temp_id_pos_vector = temp_vector > 0
        if any(temp_id_pos_vector):
            temp_pos_sum = np.sum(temp_vector[temp_id_pos_vector])
        else:
            temp_pos_sum = 0
        #
        temp_id_neg_vector = temp_vector < 0
        if any(temp_id_neg_vector):
            temp_neg_sum = np.sum(temp_vector[temp_id_neg_vector])
        else:
            temp_neg_sum = 0
        #
        temp_sum = temp_pos_sum - temp_neg_sum
        #
        if temp_sum == 0:
            continue
        #
        if temp_sum > 0:
            temp_c_value = temp_sum * temp_ceil_vector
            #
            temp_id_vector = temp_vector > temp_c_value
            if any(temp_id_vector):
                id4allok = False
                temp_vector[temp_id_vector] = temp_c_value[temp_id_vector]
            #
            temp_id_vector = temp_vector < -temp_c_value
            if any(temp_id_vector):
                id4allok = False
                temp_vector[temp_id_vector] = -temp_c_value[temp_id_vector]
            #
            if any(temp_id_pos_vector):
                temp_temp_sum = np.sum(temp_vector[temp_id_pos_vector])
                temp_vector[temp_id_pos_vector] = temp_pos_sum * temp_vector[temp_id_pos_vector] / temp_temp_sum
            if any(temp_id_neg_vector):
                temp_temp_sum = np.sum(temp_vector[temp_id_neg_vector])
                temp_vector[temp_id_neg_vector] = temp_neg_sum * temp_vector[temp_id_neg_vector] / temp_temp_sum
            #
            OpedFactorMonlix[i, :] = temp_vector

    # Iteration
    OpParaNumIteration4Truncate = OpParaNumIteration4Truncate - 1
    if id4allok:
        OpParaNumIteration4Truncate = 0

    if OpParaNumIteration4Truncate == 0:
        return OpedFactorMonlix
    #
    OpedFactorMonlix = weight_control(OpedFactorMonlix, OpParaPercent=OpParaPercent,
                                      OpParaNumIteration4Truncate=OpParaNumIteration4Truncate)
    return OpedFactorMonlix


@nb.jit(nopython=True)
def dropna_val(y_arr, mv_arr, idu_arr):
    idu_arr[np.isnan(idu_arr)] = 99887766
    mask = ~(np.isnan(y_arr) | np.isnan(mv_arr))  # | np.isnan(idu_arr))
    y_arr, mv_arr, idu_arr = y_arr[mask], mv_arr[mask], idu_arr[mask]
    y_arr, mv_arr = np.expand_dims(y_arr, axis=0), np.expand_dims(mv_arr, axis=0)
    return y_arr, mv_arr, idu_arr, mask


def neutralize(mat, sqmv, indu, mv_neutral=True, idu_neutral=True):
    if not (mv_neutral | idu_neutral):
        return mat
    print('[%s][start: factor neurtral]' % datetime.fromtimestamp(time()).strftime('%T'))
    one_arr = 100 * np.ones_like(mat[[0]])
    for i, (y_arr, mv_arr, idu_arr) in enumerate(zip(mat, sqmv, indu)):
        y_arr, mv_arr, idu_arr, mask = dropna_val(y_arr, mv_arr, idu_arr)
        x_arr = [one_arr[:, : y_arr.size] / y_arr.size]
        if mv_neutral:
            ln_mv = np.log(mv_arr)
            mv_arr = 100 * ln_mv / (np.mean(ln_mv) * ln_mv.size)
            x_arr.append(mv_arr)
        if idu_neutral:
            x_arr.append(indu_encode(idu_arr))
        x_arr = np.concatenate(x_arr, axis=0)
        try:
            mat[i, mask] = reg_residual(y_arr, x_arr)
        except LinAlgError:
            mat[i] = np.full_like(mask, np.nan)
    return mat


def mypivot(df, timeindex, instruments, index, columns, values):
    calidx = {c: i for i, c in enumerate(timeindex)}
    instidx = {inst: i for i, inst in enumerate(instruments)}

    df['calidx'] = df[index].map(calidx)
    df['instidx'] = df[columns].map(instidx)

    df = df.dropna(subset=['calidx', 'instidx'], axis=0, how='any')

    mat = np.full((timeindex.size, instruments.size), np.nan)
    mat[df.calidx.values.astype(int), df.instidx.values.astype(int)] = df[values].values
    return mat


def update_weight(position, factor, univ, tradable):
    new_weight = np.zeros_like(position)
    new_weight[tradable] = get_tgt_weight(factor[tradable], univ[tradable])

    frozen_posi = position[~tradable]
    new_weight[new_weight < 0] *= 1 + frozen_posi[frozen_posi < 0].sum()
    new_weight[new_weight > 0] *= 1 - frozen_posi[frozen_posi > 0].sum()

    new_weight[~tradable] = position[~tradable]
    turnover = np.abs(new_weight - position).sum()
    # assert np.sum(new_weight) < 0.0000000001
    # assert np.sum(new_weight) < 0.0000000001
    # assert np.sum(new_weight) < 0.0000000001
    return new_weight, turnover


def get_tgt_weight(fact_i, univ_i):
    """ cal target weight"""
    fact_i[~univ_i] = 0
    fact_i[np.isnan(fact_i)] = 0
    l_posi = fact_i > 0
    s_posi = fact_i < 0
    fact_i[l_posi] /= np.nansum(fact_i[l_posi])
    fact_i[s_posi] /= - np.nansum(fact_i[s_posi])
    return fact_i


def weightby(arr1, arr2, mmin=True):
    if mmin:
        arr2 -= arr2.min()
    arr2_sum = arr2.sum()
    if arr2_sum == 0:
        return arr1.mean()
    return np.dot(arr1, arr2) / arr2_sum


def equalweight(arr1, arr2):
    return np.nanmean(arr1)


def cal_group_return(fact_i, retr_i, ngroups):
    tmp = np.full((ngroups, ), 0, dtype=np.float64)
    rank = bn.nanrankdata(fact_i)
    nrange = np.sum(~np.isnan(rank))
    norm_rank = (ngroups * (rank - 1) / (nrange - 1)).astype(int)
    norm_rank = np.fmin(norm_rank, ngroups - 1)

    uniq, indices = np.unique(norm_rank, return_inverse=True)
    groups = np.full((uniq.shape[0], indices.shape[0]), np.nan)
    arrs = np.arange(indices.shape[0])
    groups[indices, arrs] = retr_i
    groupret = np.nanmean(groups, axis=1)
    if uniq.size == ngroups:
        return groupret
    tmp[uniq.astype(int)] = groupret
    return tmp


@nb.jit(nopython=True)
def nancorr(arr1, arr2):
    mask = np.isnan(arr1) | np.isnan(arr2)
    if mask.all():
        return np.nan
    arr1 = arr1[~mask]
    arr2 = arr2[~mask]
    arr1 = arr1 - np.mean(arr1)
    arr2 = arr2 - np.mean(arr2)
    std1 = max(np.sqrt(np.mean(arr1 * arr1)), 0.001)
    std2 = max(np.sqrt(np.mean(arr2 * arr2)), 0.001)
    return np.mean(arr1 * arr2) / (std1 * std2)


@nb.jit(nopython=True)
def Nancorr(mat1, mat2):
    nrow, _ = mat1.shape
    temp_data = np.zeros((nrow,))
    for i in np.arange(nrow):
        arr1 = mat1[i]
        arr2 = mat2[i]
        ic = nancorr(arr1, arr2)
        temp_data[i] = ic
    return temp_data


def ranknancorr(arr1, arr2):
    mask = np.isnan(arr1) | np.isnan(arr2)
    if mask.all():
        return np.nan
    arr1, arr2 = arr1[~mask], arr2[~mask]
    arr1, arr2 = bn.rankdata(arr1), bn.rankdata(arr2)
    arr1, arr2 = arr1 - np.mean(arr1), arr2 - np.mean(arr2)
    std = np.std(arr1) * np.std(arr2)
    if std == 0:
        return 0
    return np.mean(arr1 * arr2) / std


def RankNancorr(mat1, mat2):
    nrow, _ = mat1.shape
    temp_data = np.zeros((nrow,))
    for i in np.arange(nrow):
        arr1 = mat1[i]
        arr2 = mat2[i]
        ic = ranknancorr(arr1, arr2)
        temp_data[i] = ic
    return temp_data


if __name__ == '__main__':
    indu = ED.idu_info
    sqmv = ED.mv_info
    mat = ED.ret_info
    mat = neutralize(mat, sqmv, indu, True, True)
    pass



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@File    :   data_utils.py    
@Contact :   wangcc@csc.com.cn
@License :   (C)Copyright 2017-2018, CSC

@Modify Time      @Author    @Version    @Desciption
------------      -------    --------    -----------
2021/11/2 14:57   wangcc     1.0         None
'''

# import lib
import numpy as np
from BBData import D
from BBData.reader_utils import Cal, Inst
from BBData.config import cal_config
from .config import up_bound_r, dn_bound_r, markets


class EngineData:

    def __init__(self, start_date=None, end_date=None):
        self.start_date = start_date
        self.end_date = end_date
        self.H = {}

    def get_info(self, field, freq="Tday"):
        if field in self.H:
            return self.H[field]
        mat = D.features(field, self.start_date, self.end_date, freq=freq, return_mat=True)[field]
        self.H[field] = mat
        return mat

    @property
    def st_info(self):
        return np.isnan(self.get_info("ndns_st"))

    @property
    def sus_info(self):
        return np.isnan(self.get_info("ndns_sus"))

    @property
    def ipo_info(self):
        return self.get_info("ndns_ipo") == 1

    @property
    def ret_info(self):
        return self.get_info("pctchange")

    @property
    def mv_info(self):
        return self.get_info("ndns_mv")

    @property
    def idu_info(self):
        return self.get_info("ndns_indu")

    @property
    def valmv_info(self):
        return self.get_info("ndns_valmv")

    @property
    def stk_univ(self):
        return self.get_info("ndns_indu")

    def trade_status(self, stk_univ, si, ei):
        assert stk_univ in markets, "stk_univ not exist"
        if stk_univ in self.H:
            stk_univ_, tradable = self.H[stk_univ]
        else:
            stk_univ_ = self.st_info & self.ipo_info & self.sus_info
            tradable = (self.ret_info > dn_bound_r) & (self.ret_info < up_bound_r) & self.sus_info
            if stk_univ != "all":
                stk_univ_ &= (self.get_info(f"univ_mask_{stk_univ}") == 1)
            self.H[stk_univ] = (stk_univ_, tradable)
        return stk_univ_[si: ei+1], tradable[si: ei+1]


ED = EngineData()


if __name__ == '__main__':
    pass